MODULE Perfect;
(**
Build Perfect Hash Function with Chang-Lee method.
*)

IMPORT
    SYSTEM;

TYPE
    (* HashInt* = INTEGER; *)
    HashInt* = LONGINT;

    HashItem* = RECORD
        index*  : LONGINT;
        h1*, h2*: HashInt;
        rh1, rh2: LONGINT;
    END;

    HashList* = POINTER TO ARRAY OF HashItem;
    HashVector* = POINTER TO ARRAY OF HashInt;
    LIntVector* = POINTER TO ARRAY OF LONGINT;

TYPE
    CompareProc = PROCEDURE (list: SYSTEM.PTR; x, y: LONGINT): BOOLEAN;
    SwapProc = PROCEDURE (list: SYSTEM.PTR; x, y: LONGINT);

PROCEDURE QSort* (list: SYSTEM.PTR; first, last: LONGINT; Less: CompareProc; Swap: SwapProc);

    PROCEDURE Sort (l, r: LONGINT);
    VAR
        i, j: LONGINT;
    BEGIN
        WHILE r > l DO
            i := l+1;
            j := r;
            WHILE i <= j DO
                WHILE (i <= j) & ~Less(list, l, i) DO INC(i) END;
                WHILE (i <= j) & Less(list, l, j) DO DEC(j) END;
                IF i <= j THEN Swap(list, i, j); INC(i); DEC(j) END;
            END (*WHILE*);
            IF j # l THEN Swap(list, j, l) END;
            IF (j+j) > (r+l) THEN
                Sort(j+1, r);
                r := j-1;
            ELSE
                Sort(l, j-1);
                l := j+1;
            END (*IF*);
        END (*WHILE*);
    END Sort;

BEGIN
    ASSERT(first <= last);
    Sort(first, last);
END QSort;


PROCEDURE LessHashItem (l: SYSTEM.PTR; x, y: LONGINT): BOOLEAN;
VAR list: HashList;
BEGIN
    list := SYSTEM.VAL(HashList, l);

    IF list[x].h1 < list[y].h1 THEN
        RETURN TRUE;
    ELSIF list[x].h1 > list[y].h1 THEN
        RETURN FALSE;
    ELSIF list[x].h2 < list[y].h2 THEN
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END (*IF*);
END LessHashItem;

PROCEDURE SwapHashItem (l: SYSTEM.PTR; x, y: LONGINT);
VAR
    list: HashList;
    item: HashItem;
BEGIN
    list := SYSTEM.VAL(HashList, l);

    item := list[x];
    list[x] := list[y];
    list[y] := item;
END SwapHashItem;

PROCEDURE LessHashItemH1 (l: SYSTEM.PTR; x, y: LONGINT): BOOLEAN;
VAR list: HashList;
BEGIN
    list := SYSTEM.VAL(HashList, l);
    RETURN list[x].rh1 < list[y].rh1;
END LessHashItemH1;

PROCEDURE LessHashItemH2 (l: SYSTEM.PTR; x, y: LONGINT): BOOLEAN;
VAR list: HashList;
BEGIN
    list := SYSTEM.VAL(HashList, l);
    RETURN list[x].rh2 < list[y].rh2;
END LessHashItemH2;

PROCEDURE LessLInt (l: SYSTEM.PTR; x, y: LONGINT): BOOLEAN;
VAR list: LIntVector;
BEGIN
    list := SYSTEM.VAL(LIntVector, l);
    RETURN list[x] < list[y];
END LessLInt;

PROCEDURE SwapLInt (l: SYSTEM.PTR; x, y: LONGINT);
VAR
    list: LIntVector;
    item: LONGINT;
BEGIN
    list := SYSTEM.VAL(LIntVector, l);

    item := list[x];
    list[x] := list[y];
    list[y] := item;
END SwapLInt;


(* make the array of first prime numbers using Erathophen method *)
PROCEDURE MakePrimes (VAR primes: ARRAY OF HashInt);
VAR
    i, j: LONGINT;
    n   : HashInt;
BEGIN
    primes[0] := 1;
    n := 1;

    FOR i := 1 TO LEN(primes)-1 DO

        REPEAT
            INC(n);
            j := 1;
            WHILE (j < i) & (n MOD primes[j] # 0) DO INC(j) END;
        UNTIL j = i;

        primes[i] := n;

    END (*FOR*);
END MakePrimes;

(* solve "chinese remainder theorem", find c: forall i=0..n-1 (c MOD p[i]) = r[i] *)
PROCEDURE SolveChineseRemainder (VAR result: HashInt; n: LONGINT; p, r: ARRAY OF HashInt): BOOLEAN;
VAR
    i, j: LONGINT;
    mult: HashInt;
BEGIN
    ASSERT(n > 0);

    result := r[0];
    mult := p[0];

    FOR i := 1 TO n-1 DO
        ASSERT(r[i] < p[i]);

        j := 1;
        WHILE (result MOD p[i]) # r[i] DO

            IF j = p[i] THEN
                (* no result found on the i-th step *)
                RETURN FALSE;
            END (*IF*);

            IF result > (MAX(HashInt) - mult) THEN
                (* overflow in HashInt *)
                RETURN FALSE;
            END (*IF*);

            INC(result, mult);
            INC(j);
        END (*WHILE*);

        IF mult > (MAX(HashInt) DIV p[i]) THEN
            (* overflow in HashInt *)
            RETURN FALSE;
        END (*IF*);
        mult := mult * p[i];
    END (*FOR*);

    RETURN TRUE;
END SolveChineseRemainder;


PROCEDURE MakeHash* (list: HashList; VAR d, c, p: HashVector): BOOLEAN;
CONST
    primesNumber = 55;

VAR
    j: HashInt;
    i, n, mMax: LONGINT;
    ln, lm: LONGINT;
    h1, h2: LONGINT;
    values: LIntVector;
    primes, remainders: ARRAY primesNumber OF HashInt;

    PROCEDURE TryHash2 (): BOOLEAN;
    VAR
        i, j, l, n, rh1: LONGINT;
        p: ARRAY primesNumber OF HashInt;
    BEGIN
        n := LEN(list^);
    
        FOR i := 0 TO n-1 DO
            list[i].rh2 := list[i].h2 MOD ln;
        END (*FOR*);
    
        i := 0;
        WHILE i < n DO
            l := i;
            rh1 := list[i].rh1;
    
            (* looking for hash slot of words with the same rh1 *)
            REPEAT INC(i) UNTIL (i = n) OR (list[i].rh1 # rh1);
    
            QSort(list, l, i-1, LessHashItemH2, SwapHashItem);
    
            FOR j := 0 TO i-l-1 DO
                p[j] := primes[list[j+l].rh2];
    
                (* incorrect condition for "chinese remainder theorem" *)
                IF remainders[j] >= p[j] THEN
                    RETURN FALSE;
                END (*IF*);
            END (*FOR*);
    
            (* solve "chinese remainder theorem" *)
            IF ~SolveChineseRemainder(c[rh1], i-l, p, remainders) THEN
                RETURN FALSE;
            END (*IF*);
        END (*WHILE*);
    
        RETURN TRUE;
    END TryHash2;
    
    PROCEDURE TryHash (): BOOLEAN;
    VAR
        i, l, n, nMax, rh1: LONGINT;
        rd: HashInt;
    BEGIN
        n := LEN(list^);
    
        lm := n DIV primesNumber;
        IF lm < 1 THEN lm := 1 END;
    
        (* trying to build hash function for different h1(w) = H1(w) MOD lm *)
        WHILE lm < mMax DO
            FOR i := 0 TO n-1 DO
                list[i].rh1 := list[i].h1 MOD lm;
            END (*FOR*);
    
            QSort(list, 0, n-1, LessHashItemH1, SwapHashItem);
    
            (* find nMax - maximal number of words in one h1() hash slot *)
            nMax := 0;
            rd   := 0;
    
            i := 0;
            WHILE i < n DO
                l := i;
                rh1 := list[i].rh1;
    
                REPEAT INC(i) UNTIL (i = n) OR (list[i].rh1 # rh1);
    
                d[rh1] := rd;
                INC(rd, i-l);
                IF nMax < i-l THEN nMax := i-l END;
            END (*WHILE*);
    
            (* found correct result for this case *)
            IF nMax < primesNumber THEN
                (* trying different h2(w) = H2(w) MOD ln *)
                ln := nMax;
                WHILE ln <= primesNumber DO
                    IF TryHash2() THEN
                        RETURN TRUE;
                    END (*IF*);
                    INC(ln);
                END (*WHILE*);
            END (*IF*);
    
            INC(lm);
        END (*WHILE*);
    
        RETURN FALSE;
    END TryHash;

    PROCEDURE Copy (VAR src: ARRAY OF HashInt; n: LONGINT): HashVector;
    VAR
        dest: HashVector;
        i   : LONGINT;
    BEGIN
        NEW(dest, n);
    
        FOR i := 0 TO n-1 DO
            dest[i] := src[i];
        END (*FOR*);

        RETURN dest;
    END Copy;
    
BEGIN
    n := LEN(list^);
    ASSERT(n > 1);

    (* be sure, that all hash values pairs are different *)
    QSort(list, 0, n-1, LessHashItem, SwapHashItem);

    FOR i := 0 TO n-2 DO
        IF (list[i].h1 = list[i+1].h1) & (list[i].h2 = list[i+1].h2) THEN
            RETURN FALSE;
        END (*IF*);
    END (*FOR*);

    (* fill the primes array *)
    MakePrimes(primes);

    (* fill the remainders array *)
    FOR j := 0 TO primesNumber-1 DO
        remainders[j] := j;
    END (*FOR*);

    (* trying to build hash *)
    mMax := n*5;

    NEW(d, mMax);
    NEW(c, mMax);

    IF ~TryHash() THEN
        RETURN FALSE;
    END (*IF*);

    (* produce the result *)
    p := Copy(primes, ln);

    c := Copy(c^, lm);
    d := Copy(d^, lm);

    (* verify the result *)
    NEW(values, n);

    FOR i := 0 TO n-1 DO
        h1 := list[i].h1 MOD lm;
        h2 := list[i].h2 MOD ln;
        values[i] := d[h1] + c[h1] MOD p[h2];
    END (*FOR*);

    QSort(values, 0, n-1, LessLInt, SwapLInt);

    FOR i := 0 TO n-1 DO
        IF values[i] # i THEN
            RETURN FALSE;
        END (*IF*);
    END (*FOR*);

    RETURN TRUE;
END MakeHash;


END Perfect.
